#include <stdint.h>
#include <toolchain.h>

#define CPUCLOCK    73728000    /* CPU clock in Hz */
#define BAUDRATE    115200      /* UART baud rate */

/*
 * Spurious interrupt entry.
 */
static void SpuriousEntry(void) __attribute__ ((naked));
void SpuriousEntry(void)
{
    IRQ_ENTRY();
    IRQ_EXIT();
}

/*
 * Calculate UART baud rate divider.
 */
static uint32_t At91BaudRateDiv(uint32_t baud)
{
    /* See datasheet. */
    return (CPUCLOCK / (8 * baud) + 1) / 2;
}

/*
 * Output single character at UART0.
 */
static void PutChar(char ch)
{
    /* Prepend linefeed with carriage return. */
    if (ch == '\n') {
        PutChar('\r');
    }
    /* Wait until transmitter ready. */
    while ((inr(US0_CSR) & US_TXRDY) == 0);
    /* Send the next character. */
    outr(US0_THR, ch);
}

/*
 * Output string at UART0.
 */
static void PutString(char *str)
{
    /* Loop for all characters until the end of the string. */
    while (*str) {
        PutChar(*str);
        str++;
    }
}

/*
 * The Nut/OS runtime initialization calls NutInit() instead of main().
 */
void NutInit(void)
{
    /* Set spurious interrupt vector. */
    outr(AIC_SPU, (unsigned int)SpuriousEntry);
    /* Enable UART0 clock. */
    outr(PS_PCER, _BV(US0_ID));
    /* Disable peripheral function at UART0 transmit pin. */
    outr(PIO_PDR, _BV(P14_TXD0));
    /* Reset UART. */
    outr(US0_CR, US_RSTRX | US_RSTTX | US_RXDIS | US_TXDIS);
    /* Disable all UART interrupts. */
    outr(US0_IDR, 0xFFFFFFFF);
    /* Set UART baud rate generator register. */
    outr(US0_BRGR, At91BaudRateDiv(BAUDRATE));
    /* Set UART mode to 8 data bits, no parity and 1 stop bit. */
    outr(US0_MR, US_CHMODE_NORMAL | US_CHRL_8 | US_PAR_NO | US_NBSTOP_1);
    /* Enable UART transmitter. */
    outr(US0_CR, US_TXEN);
    /* Print message in a loop. */
    for (;;) {
        PutString("Hello world!\n");
    }
}
